@@ -3,7 +3,6 @@ from keras import backend as K
 from keras.layers import Concatenate
 from keras.layers import Input, Dense, Lambda, Subtract, Add, Reshape
 from keras.models import Model
-from keras.optimizers import Adam
 
 
 class NBeatsNet:
@@ -11,6 +10,9 @@ class NBeatsNet:
     TREND_BLOCK = 'trend'
     SEASONALITY_BLOCK = 'seasonality'
 
+    _BACKCAST = 'backcast'
+    _FORECAST = 'forecast'
+
     def __init__(self,
                  input_dim=1,
                  exo_dim=0,
@@ -21,8 +23,7 @@ class NBeatsNet:
                  thetas_dim=(4, 8),
                  share_weights_in_stack=False,
                  hidden_layer_units=256,
-                 nb_harmonics=None
-                 ):
+                 nb_harmonics=None):
 
         self.stack_types = stack_types
         self.nb_blocks_per_stack = nb_blocks_per_stack
@@ -67,21 +68,27 @@ class NBeatsNet:
 
         for k in range(self.input_dim):
             y_[k] = Reshape(target_shape=(self.forecast_length, 1))(y_[k])
+            x_[k] = Reshape(target_shape=(self.backcast_length, 1))(x_[k])
         if self.input_dim > 1:
-            y_ = Concatenate(axis=-1)([y_[ll] for ll in range(self.input_dim)])
+            y_ = Concatenate()([y_[ll] for ll in range(self.input_dim)])
+            x_ = Concatenate()([x_[ll] for ll in range(self.input_dim)])
         else:
             y_ = y_[0]
+            x_ = x_[0]
 
         if self.has_exog():
-            model = Model([x, e], y_)
+            n_beats_forecast = Model([x, e], y_, name=self._FORECAST)
+            n_beats_backcast = Model([x, e], x_, name=self._BACKCAST)
         else:
-            model = Model(x, y_)
+            n_beats_forecast = Model(x, y_, name=self._FORECAST)
+            n_beats_backcast = Model(x, x_, name=self._BACKCAST)
 
-        self.n_beats = model
+        self.models = {model.name: model for model in [n_beats_backcast, n_beats_forecast]}
+        self.cast_type = self._FORECAST
 
     def has_exog(self):
-        # exo/exog is short for "exogenous variable", i.e. any input
-        # features other than the target timeseries itself.
+        # exo/exog is short for 'exogenous variable', i.e. any input
+        # features other than the target time-series itself.
         return self.exo_dim > 0
 
     @staticmethod
@@ -105,7 +112,6 @@ class NBeatsNet:
         return layer_with_weights
 
     def create_block(self, x, e, stack_id, block_id, stack_type, nb_poly):
-
         # register weights (useful when share_weights_in_stack=True)
         def reg(layer):
             return self._r(layer, stack_id)
@@ -127,10 +133,10 @@ class NBeatsNet:
             forecast = reg(Dense(self.forecast_length, activation='linear', name=n('forecast')))
         elif stack_type == 'trend':
             theta_f = theta_b = reg(Dense(nb_poly, activation='linear', use_bias=False, name=n('theta_f_b')))
-            backcast = Lambda(trend_model, arguments={"is_forecast": False, "backcast_length": self.backcast_length,
-                                                      "forecast_length": self.forecast_length})
-            forecast = Lambda(trend_model, arguments={"is_forecast": True, "backcast_length": self.backcast_length,
-                                                      "forecast_length": self.forecast_length})
+            backcast = Lambda(trend_model, arguments={'is_forecast': False, 'backcast_length': self.backcast_length,
+                                                      'forecast_length': self.forecast_length})
+            forecast = Lambda(trend_model, arguments={'is_forecast': True, 'backcast_length': self.backcast_length,
+                                                      'forecast_length': self.forecast_length})
         else:  # 'seasonality'
             if self.nb_harmonics:
                 theta_b = reg(Dense(self.nb_harmonics, activation='linear', use_bias=False, name=n('theta_b')))
@@ -138,11 +144,11 @@ class NBeatsNet:
                 theta_b = reg(Dense(self.forecast_length, activation='linear', use_bias=False, name=n('theta_b')))
             theta_f = reg(Dense(self.forecast_length, activation='linear', use_bias=False, name=n('theta_f')))
             backcast = Lambda(seasonality_model,
-                              arguments={"is_forecast": False, "backcast_length": self.backcast_length,
-                                         "forecast_length": self.forecast_length})
+                              arguments={'is_forecast': False, 'backcast_length': self.backcast_length,
+                                         'forecast_length': self.forecast_length})
             forecast = Lambda(seasonality_model,
-                              arguments={"is_forecast": True, "backcast_length": self.backcast_length,
-                                         "forecast_length": self.forecast_length})
+                              arguments={'is_forecast': True, 'backcast_length': self.backcast_length,
+                                         'forecast_length': self.forecast_length})
         for k in range(self.input_dim):
             if self.has_exog():
                 d0 = Concatenate()([x[k]] + [e[ll] for ll in range(self.exo_dim)])
@@ -159,40 +165,36 @@ class NBeatsNet:
 
         return backcast_, forecast_
 
-    def compile_model(self, loss, learning_rate):
-        optimizer = Adam(lr=learning_rate)
-        self.compile(loss=loss, optimizer=optimizer)
-
     def __getattr__(self, name):
         # https://github.com/faif/python-patterns
         # model.predict() instead of model.n_beats.predict()
         # same for fit(), train_on_batch()...
-        attr = getattr(self.n_beats, name)
+        attr = getattr(self.models[self._FORECAST], name)
 
         if not callable(attr):
             return attr
 
         def wrapper(*args, **kwargs):
-            return attr(*args, **kwargs)
+            cast_type = self._FORECAST
+            if attr.__name__ == 'predict' and 'return_backcast' in kwargs and kwargs['return_backcast']:
+                del kwargs['return_backcast']
+                cast_type = self._BACKCAST
+            return getattr(self.models[cast_type], attr.__name__)(*args, **kwargs)
 
         return wrapper
 
 
-def linear_space(backcast_length, forecast_length, fwd_looking=True):
-    ls = K.arange(-float(backcast_length), float(forecast_length), 1) / backcast_length
-    if fwd_looking:
-        ls = ls[backcast_length:]
-    else:
-        ls = ls[:backcast_length]
-    return ls
+def linear_space(backcast_length, forecast_length, is_forecast=True):
+    ls = K.arange(-float(backcast_length), float(forecast_length), 1) / forecast_length
+    return ls[backcast_length:] if is_forecast else ls[:backcast_length]
 
 
 def seasonality_model(thetas, backcast_length, forecast_length, is_forecast):
     p = thetas.get_shape().as_list()[-1]
     p1, p2 = (p // 2, p // 2) if p % 2 == 0 else (p // 2, p // 2 + 1)
-    t = linear_space(backcast_length, forecast_length, fwd_looking=is_forecast)
-    s1 = K.stack([K.cos(2 * np.pi * i * t) for i in range(p1)], axis=0)
-    s2 = K.stack([K.sin(2 * np.pi * i * t) for i in range(p2)], axis=0)
+    t = linear_space(backcast_length, forecast_length, is_forecast=is_forecast)
+    s1 = K.stack([K.cos(2 * np.pi * i * t) for i in range(p1)])
+    s2 = K.stack([K.sin(2 * np.pi * i * t) for i in range(p2)])
     if p == 1:
         s = s2
     else:
@@ -203,7 +205,7 @@ def seasonality_model(thetas, backcast_length, forecast_length, is_forecast):
 
 def trend_model(thetas, backcast_length, forecast_length, is_forecast):
     p = thetas.shape[-1]
-    t = linear_space(backcast_length, forecast_length, fwd_looking=is_forecast)
-    t = K.transpose(K.stack([t ** i for i in range(p)], axis=0))
+    t = linear_space(backcast_length, forecast_length, is_forecast=is_forecast)
+    t = K.transpose(K.stack([t ** i for i in range(p)]))
     t = K.cast(t, np.float32)
     return K.dot(thetas, K.transpose(t))
